!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Mathematical and Computational Sciences Division                !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module basis_at_quad_pts

!----------------------------------------------------
! This module contains routines that return the basis functions at the
! quadrature points.
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
use message_passing, only : fatal
use quadrature_rules
use basis_functions
!----------------------------------------------------

implicit none
private
public basis_at_quad_pts_tri_nqpoints, p_hier_basis_at_quad_pts_tri, &
       basis_at_quad_pts_tet_nqpoints, p_hier_basis_at_quad_pts_tet

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------
! The following defined constants are defined:

!----------------------------------------------------
! The following types are defined:

type basis_tables_type
   real(my_real), allocatable :: basis(:,:), basisx(:,:), basisy(:,:), &
                                 basisz(:,:)
   integer :: degree=-1
end type basis_tables_type

type quadrature_point_permutation_edge
   integer, allocatable :: perm(:,:)
end type quadrature_point_permutation_edge

type quadrature_point_permutation_face
   integer, allocatable :: perm(:,:,:)
end type quadrature_point_permutation_face

!----------------------------------------------------
! The following variables are defined:

! The bases are all p-hierarchical; if I decide to do other basis types, these
! arrays will need a second dimension for the degree.
!
! The i'th entry contains a table of basis functions and derivatives for
! the quadrature rule of order i.  The table will be dimensioned and filled
! to hold the basis of the highest degree seen so far for that quadrature rule.
! Because the basis is p-hierarchical, if a lower degree is requested it can be
! extracted from the table.

type(basis_tables_type) :: basis_table_tri(MAX_QUAD_ORDER_TRI)
type(basis_tables_type) :: basis_table_tet(MAX_QUAD_ORDER_TET)

type(quadrature_point_permutation_edge) :: qp_perm_edge(MAX_QUAD_ORDER_TET)
type(quadrature_point_permutation_face) :: qp_perm_face(MAX_QUAD_ORDER_TET)

!----------------------------------------------------

contains

!        ------------------------------
function basis_at_quad_pts_tri_nqpoints(order)
!        ------------------------------

!----------------------------------------------------
! This routine returns the number of quadrature points for the triangle
! quadrature rule of order order.  It can be used to determine the second
! dimension of basis, etc., in subroutine *_basis_at_quad_pts_tri.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: order
integer :: basis_at_quad_pts_tri_nqpoints

!----------------------------------------------------
! Local variables:

integer :: i
integer, save :: nqpoints(MAX_QUAD_ORDER_TRI) = (/(0, i=1,MAX_QUAD_ORDER_TRI)/)
integer :: nq, jerr
real(my_real), pointer :: qweights(:),xquad(:),yquad(:)
real(my_real) :: xvert(3) = (/0.0_my_real,0.0_my_real,1.0_my_real/), &
                 yvert(3) = (/0.0_my_real,1.0_my_real,0.0_my_real/)
!----------------------------------------------------
! Begin executable code

if (order < 1 .or. order > MAX_QUAD_ORDER_TRI) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("basis_at_quad_pts_tri_nqpoints: order out of range", &
              intlist = (/order/))
   stop
elseif (nqpoints(order) == 0) then
   call quadrature_rule_tri(order,xvert,yvert,nq,qweights,xquad,yquad, &
                            jerr,stay_in=.true.)
   if (jerr /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("basis_at_quad_pts_tri_nqpoints: quadrature_rule_tri returned error code", &
                 intlist = (/jerr/))
      stop
   else
      deallocate(qweights,xquad,yquad)
      nqpoints(order) = nq
      basis_at_quad_pts_tri_nqpoints = nq
   endif
else
   basis_at_quad_pts_tri_nqpoints = nqpoints(order)
endif

end function basis_at_quad_pts_tri_nqpoints

!          ----------------------------
subroutine p_hier_basis_at_quad_pts_tri(order,degree,xvert,yvert,basis,basisx, &
                                        basisy,nqpoints,qweights,xquad, &
                                        yquad)
!          ----------------------------

!----------------------------------------------------
! This routine returns the basis functions at quadrature points of a triangle
! in the plane.  order is the order of the quadrature rule.  degree(1:3) is the
! degree of the edge bases opposite the corresponding vertex; degree(4) is the
! degree of the bubble bases.  basis, basisx and basisy are all optional.
! xvert and yvert are the vertices of the triangle.  nqpoints, qweights, xquad
! and yquad are all optional, and return the quadrature rule if present.  If one
! of them is present, they must all be present.  There is (currently) no
! provision for returning the red or black subset of bases -- all bases are
! returned.  This routine always uses stay_in=.true.
!
! basis(i,j) (and the derivatives) is the i'th basis function at the j'th
! quadrature point.  The order of the basis functions is the same as in module
! basis_functions, and the order of the quadrature points is the same as in
! module quadrature_rules.  As in module quadrature rules, qweights, xquad and
! yquad are pointers which are allocated here and must be deallocated by the
! caller.  As in module basis functions, basis, basisx and basisy must be
! correctly dimensioned in the caller.  The first dimension (number of basis
! functions) is the same as when allocating for module basis functions.  The
! second dimension (number of quadrature points) can be obtained from function
! basis_at_quad_pts_tri_nqpoints.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: order, degree(4)
real(my_real), intent(in) :: xvert(:), yvert(:)
real(my_real), optional, intent(out) :: basis(:,:), basisx(:,:), basisy(:,:)
integer, optional, intent(out) :: nqpoints
real(my_real), optional, pointer :: qweights(:), xquad(:), yquad(:)
!----------------------------------------------------
! Local variables:

integer :: nq, astat, jerr, deg, nbasis, i, isub, &
           edge, jstart, jnum
integer, allocatable :: subtable(:)
logical :: needA
real(my_real), pointer :: qw(:), xq(:), yq(:)
real(my_real), allocatable :: b(:,:), bx(:,:), by(:,:)
real(my_real) :: A(2,2), denom
real(my_real) :: xv(3) = (/0.0_my_real,0.0_my_real,1.0_my_real/), &
                 yv(3) = (/0.0_my_real,1.0_my_real,0.0_my_real/)
!----------------------------------------------------
! Begin executable code

! Check correctness of optional argument combinations

if ((present(nqpoints) .neqv. present(qweights)) .or. &
    (present(nqpoints) .neqv. present(xquad)) .or. &
    (present(nqpoints) .neqv. present(yquad))) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("p_hier_basis_at_quad_pts_tri: all or none of the quadrature rule arguments must be present")
   stop
endif

if (order < 1 .or. order > MAX_QUAD_ORDER_TRI) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("p_hier_basis_at_quad_pts_tri: order out of range")
   stop
endif

nullify(qw) ! So I can test if the quadrature rule has been set

! Get basis functions for the maximum degree of this element if they have not
! already been cached for this quadrature order.

! For OpenMP, only one thread should allocate basis_table_tri, so make this
! critical after the test (so it is not critical all the time), and then do
! the test again, which should fail if the first thread made the table big
! enough.

deg = maxval(degree)
if (basis_table_tri(order)%degree < deg) then
!$omp critical
   if (basis_table_tri(order)%degree < deg) then
      call quadrature_rule_tri(order,xv,yv,nq,qw,xq,yq,jerr,stay_in=.true.)
      if (jerr /= 0) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("p_hier_basis_at_quad_pts_tri: quadrature_rule_tri returned error code",intlist=(/jerr/))
         stop
      endif
      nbasis = 3 + 3*(deg-1) + ((deg-1)*(deg-2))/2
      allocate(b(nbasis,nq),bx(nbasis,nq),by(nbasis,nq),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("p_hier_basis_at_quad_pts_tri: allocation failed")
         stop
      endif
      call p_hier_basis_func(xq,yq,xv,yv,(/deg,deg,deg,deg/),"a",b,bx,by)
      if (basis_table_tri(order)%degree > 0) then
         deallocate(basis_table_tri(order)%basis,  &
                    basis_table_tri(order)%basisx, &
                    basis_table_tri(order)%basisy)
      endif
      allocate(basis_table_tri(order)%basis (nbasis,nq), &
               basis_table_tri(order)%basisx(nbasis,nq), &
               basis_table_tri(order)%basisy(nbasis,nq), stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("p_hier_basis_at_quad_pts_tri: allocation failed")
         stop
      endif
      basis_table_tri(order)%basis  = b
      basis_table_tri(order)%basisx = bx
      basis_table_tri(order)%basisy = by
      basis_table_tri(order)%degree = deg
      deallocate(b,bx,by,qw,xq,yq)
   endif
!$omp end critical
endif

! Using the canonical triangle (0,0),(0,1),(1,0).  The transformation from a
! general triangle (x1,y1),(x2,y2),(x3,y3) to the canonical triangle in
! (xi,eta) coordinates, with (x1,y1)->(0,0), (x2,y2)->(0,1) and (x3,y3)->(1,0)
! is
!    (xi ) = A(x) + b
!    (eta)    (y)
!
! with A and b defined by
!
! A^-1 = [x3-x1 x2-x1]  and b = -A(x1)
!        [y3-y1 y2-y1]            (y1)
!
! This gives
!
! A = 1/d [ y2-y1 x1-x2 ]  b = 1/d (x2*y1-x1*y2)
!         [ y1-y3 x3-x1 ]          (x1*y3-x3*y1)
!
! with d = (x3-x1)*(y2-y1) - (y3-y1)*(x2-x1)
!
! Basis functions at the quadrature points are invariant to this transformation.
! But the derivatives are not.  For example the x derivative of f is
! df/dx = df/dxi dxi/dx + df/deta deta/dx = a_11 df/dxi + a_21 df/deta.
! Note we never use b, and we only need A if derivatives are requested.

needA = present(basisx) .or. present(basisy)

if (needA) then

   denom = (xvert(3)-xvert(1))*(yvert(2)-yvert(1)) - &
           (yvert(3)-yvert(1))*(xvert(2)-xvert(1))

   A(1,1) = (yvert(2)-yvert(1))/denom
   A(1,2) = (xvert(1)-xvert(2))/denom
   A(2,1) = (yvert(1)-yvert(3))/denom
   A(2,2) = (xvert(3)-xvert(1))/denom

endif

! If all of degree is the same as the table degree, then we don't need to
! extract a subset of bases, just transform the derivatives and copy.

if (all(degree == basis_table_tri(order)%degree)) then

   if (present(basis)) then
      basis = basis_table_tri(order)%basis
   endif

   if (present(basisx)) then
      basisx = A(1,1)*basis_table_tri(order)%basisx + &
               A(2,1)*basis_table_tri(order)%basisy
   endif

   if (present(basisy)) then
      basisy = A(1,2)*basis_table_tri(order)%basisx + &
               A(2,2)*basis_table_tri(order)%basisy
   endif

else

! Determine what subset of the table will be extracted for the given degrees

   deg = basis_table_tri(order)%degree
   allocate(subtable(size(basis,dim=1)),stat=astat)

   subtable(1:3) = (/1,2,3/)

   isub = 4
   do edge=1,3
      if (degree(edge) > 1) then
         jstart = 4 + (edge-1)*(deg-1)
         jnum = degree(edge)-1
         subtable(isub:isub+jnum-1) = (/(i,i=jstart,jstart+jnum-1)/)
         isub = isub+jnum
      endif
   end do

   if (degree(4) > 2) then
      jstart = 3*deg+1
      jnum = ((degree(4)-1)*(degree(4)-2))/2
      subtable(isub:isub+jnum-1) = (/(i,i=jstart,jstart+jnum-1)/)
      isub = isub+jnum
   endif

   nbasis = isub-1

! Copy the basis functions to the return variables, performing the
! transformations on the derivatives.

   if (present(basis)) then
      basis = basis_table_tri(order)%basis(subtable(1:nbasis),:)
   endif

   if (present(basisx)) then
      basisx = A(1,1)*basis_table_tri(order)%basisx(subtable(1:nbasis),:) + &
               A(2,1)*basis_table_tri(order)%basisy(subtable(1:nbasis),:)
   endif

   if (present(basisy)) then
      basisy = A(1,2)*basis_table_tri(order)%basisx(subtable(1:nbasis),:) + &
               A(2,2)*basis_table_tri(order)%basisy(subtable(1:nbasis),:)
   endif

   deallocate(subtable)

endif

! Return the quadrature rule if requested

if (present(nqpoints)) then
   call quadrature_rule_tri(order,xvert,yvert,nq,qw,xq,yq,jerr,stay_in=.true.)
   if (jerr /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("p_hier_basis_at_quad_pts_tri: quadrature_rule_tri returned error code",intlist=(/jerr/))
      stop
   endif
   allocate(qweights(nq),xquad(nq),yquad(nq),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("p_hier_basis_at_quad_pts_tri: allocation failed")
      stop
   endif
   nqpoints = nq
   qweights = qw
   xquad = xq
   yquad = yq
   deallocate(qw,xq,yq)
endif


end subroutine p_hier_basis_at_quad_pts_tri

!        ------------------------------
function basis_at_quad_pts_tet_nqpoints(order)
!        ------------------------------

!----------------------------------------------------
! This routine returns the number of quadrature points for the tetrahedron
! quadrature rule of order order.  It can be used to determine the second
! dimension of basis, etc., in subroutine *_basis_at_quad_pts_tet.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: order
integer :: basis_at_quad_pts_tet_nqpoints

!----------------------------------------------------
! Local variables:

integer :: i
integer, save :: nqpoints(MAX_QUAD_ORDER_TET) = (/(0, i=1,MAX_QUAD_ORDER_TET)/)
integer :: nq, jerr
real(my_real), pointer :: qweights(:),xquad(:),yquad(:),zquad(:)
real(my_real) :: xvert(4)=(/0.0_my_real,1.0_my_real,0.0_my_real,0.0_my_real/), &
                 yvert(4)=(/0.0_my_real,0.0_my_real,1.0_my_real,0.0_my_real/), &
                 zvert(4)=(/0.0_my_real,0.0_my_real,0.0_my_real,1.0_my_real/)
!----------------------------------------------------
! Begin executable code

if (order < 1 .or. order > MAX_QUAD_ORDER_TET) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("basis_at_quad_pts_tet_nqpoints: order out of range", &
              intlist = (/order/))
   stop
elseif (nqpoints(order) == 0) then
   call quadrature_rule_tet(order,xvert,yvert,zvert,nq,qweights,xquad,yquad, &
                            zquad,jerr)
   if (jerr /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("basis_at_quad_pts_tet_nqpoints: quadrature_rule_tet returned error code", &
                 intlist = (/jerr/))
      stop
   else
      deallocate(qweights,xquad,yquad,zquad)
      nqpoints(order) = nq
      basis_at_quad_pts_tet_nqpoints = nq
   endif
else
   basis_at_quad_pts_tet_nqpoints = nqpoints(order)
endif

end function basis_at_quad_pts_tet_nqpoints

!          ----------------------------
subroutine p_hier_basis_at_quad_pts_tet(order,degree,xvert,yvert,zvert,basis, &
                                        basisx,basisy,basisz,nqpoints, &
                                        qweights,xquad,yquad,zquad)
!          ----------------------------

!----------------------------------------------------
! This routine returns the basis functions at quadrature points of a
! tetrahedron.  order is the order of the quadrature rule. As in basis.f90,
! degree(1:6) is the degree of the edge bases in the canonical order w.r.t
!             the vertices: (2,3) (1,3) (1,2) (1,4) (2,4) (3,4)
! degree(7:10) is the degree of the face bases opposite vertex(1:4)
! degree(11) is the degree of the bubble bases.
! basis, basisx, basisy and basisz are all optional.
! xvert, yvert and zvert are the vertices of the triangle.  nqpoints, qweights,
! xquad, yquad and zquad are all optional, and return the quadrature rule if
! present.  If one of them is present, they must all be present.
! There is (currently) no provision for returning the red or black subset of
! bases -- all bases are returned.
!
! basis(i,j) (and the derivatives) is the i'th basis function at the j'th
! quadrature point.  The order of the basis functions is the same as in module
! basis_functions, and the order of the quadrature points is the same as in
! module quadrature_rules.  As in module quadrature rules, qweights, xquad,
! yquad and zquad are pointers which are allocated here and must be deallocated
! by the caller.  As in module basis functions, basis, basisx, basisy and basisz
! must be correctly dimensioned in the caller.  The first dimension (number of
! basis functions) is the same as when allocating for module basis functions.
! The second dimension (number of quadrature points) can be obtained from
! function basis_at_quad_pts_tet_nqpoints.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: order, degree(11)
real(my_real), intent(in) :: xvert(:), yvert(:), zvert(:)
real(my_real), optional, intent(out) :: basis(:,:), basisx(:,:), basisy(:,:), &
                                        basisz(:,:)
integer, optional, intent(out) :: nqpoints
real(my_real), optional, pointer :: qweights(:), xquad(:), yquad(:), zquad(:)
!----------------------------------------------------
! Local variables:

integer :: nq, astat, jerr, deg, nbasis, jstart, jnum, isub, edge, face, &
           alldeg(11), v1, v2, v3, v4, swapped_face_vert
logical :: needA, swapped_edge_vert
real(my_real), pointer :: qw(:), xq(:), yq(:), zq(:)
real(my_real), allocatable :: b(:,:), bx(:,:), by(:,:), bz(:,:)
real(my_real) :: Ainv(3,3), A(3,3), denom
real(my_real) :: xv(4) = (/0.0_my_real,0.0_my_real,0.0_my_real,1.0_my_real/), &
                 yv(4) = (/0.0_my_real,0.0_my_real,1.0_my_real,0.0_my_real/), &
                 zv(4) = (/0.0_my_real,1.0_my_real,0.0_my_real,0.0_my_real/)
!----------------------------------------------------
! Begin executable code

! Check correctness of optional argument combinations

if ((present(nqpoints) .neqv. present(qweights)) .or. &
    (present(nqpoints) .neqv. present(xquad)) .or. &
    (present(nqpoints) .neqv. present(yquad)) .or. &
    (present(nqpoints) .neqv. present(zquad))) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("p_hier_basis_at_quad_pts_tet: all or none of the quadrature rule arguments must be present")
   stop
endif

if (order < 1 .or. order > MAX_QUAD_ORDER_TET) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("p_hier_basis_at_quad_pts_tet: order out of range")
   stop
endif

nullify(qw) ! So I can test if the quadrature rule has been set

! Get basis functions for the maximum degree of this element if they have not
! already been cached for this quadrature order.

! For OpenMP, only one thread should allocate basis_table_tet, so make this
! critical after the test (so it is not critical all the time), and then do
! the test again, which should fail if the first thread made the table big
! enough.

deg = maxval(degree)
if (basis_table_tet(order)%degree < deg) then
!$omp critical
   if (basis_table_tet(order)%degree < deg) then
      call quadrature_rule_tet(order,xv,yv,zv,nq,qw,xq,yq,zq,jerr)
      if (jerr /= 0) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("p_hier_basis_at_quad_pts_tet: quadrature_rule_tet returned error code",intlist=(/jerr/))
         stop
      endif
      if (.not. allocated(qp_perm_edge(order)%perm)) then
         call set_quadrature_point_permutation(order)
      endif
      nbasis = 4+6*(deg-1) + 4*((deg-1)*(deg-2))/2 + ((deg-1)*(deg-2)*(deg-3))/6
      allocate(b(nbasis,nq),bx(nbasis,nq),by(nbasis,nq),bz(nbasis,nq), &
               stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("p_hier_basis_at_quad_pts_tet: allocation failed")
         stop
      endif
      alldeg = deg
      call p_hier_basis_func(xq,yq,zq,xv,yv,zv,alldeg,"a",b,bx,by,bz)
      if (basis_table_tet(order)%degree > 0) then
         deallocate(basis_table_tet(order)%basis, &
                    basis_table_tet(order)%basisx, &
                    basis_table_tet(order)%basisy, &
                    basis_table_tet(order)%basisz)
      endif
      allocate(basis_table_tet(order)%basis(nbasis,nq), &
               basis_table_tet(order)%basisx(nbasis,nq), &
               basis_table_tet(order)%basisy(nbasis,nq), &
               basis_table_tet(order)%basisz(nbasis,nq), stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("p_hier_basis_at_quad_pts_tet: allocation failed")
         stop
      endif
      basis_table_tet(order)%basis  = b
      basis_table_tet(order)%basisx = bx
      basis_table_tet(order)%basisy = by
      basis_table_tet(order)%basisz = bz
      basis_table_tet(order)%degree = deg
      deallocate(b,bx,by,bz,qw,xq,yq,zq)
   endif
!$omp end critical
endif

! Using the canonical tetrahedron (0,0,0),(0,0,1),(0,1,0),(1,0,0).
! The transformation from a general tetrahedron (x1,y1),(x2,y2),(x3,y3),(x4,y4)
! to the canonical triangle in (xi,eta,zeta) coordinates, with
! (x1,y1,z1)->(0,0,0), (x2,y2,z2)->(1,0,0), (x3,y3,z3)->(0,1,0) and
! (x4,y4,z4)->(0,0,1) is
!    (xi  )    (x)
!    (eta ) = A(y) + b
!    (zeta)    (z)
!
! with A and B defined by
!
!        [x4-x1 x3-x1 x2-x1]            (x1)
! A^-1 = |y4-y1 y3-y1 y2-y1|  and b = -A(y1)
!        [z4-z1 z3-z1 z2-z1]            (z1)
!
! Basis functions at the quadrature points are invariant to this transformation.
! But the derivatives are not.  For example the x derivative of f is
! df/dx = df/dxi dxi/dx + df/deta deta/dx + df/dzeta dzeta/dx = 
! a_11 df/dxi + a_21 df/deta + a_31 df/dzeta.
! Note we never use b, and we only need A if derivatives are requested.

needA = present(basisx) .or. present(basisy) .or. present(basisz)

! With tetrahedra there is a further complication in the transformation.
! To insure continuity across element boundaries, the vertices of an edge or
! face are sorted so they are in the same order when defining basis functions
! for adjacent elements.  This has two major effects on getting the 3D basis
! from a table.  First, the quadrature points must be permuted using the same
! reordering of points used by the basis.  This only works because the
! quadrature points are symmetric so that if, say, vertices v1 and v2 were
! swapped, I can get the correct value of the basis at a quadrature point by
! getting it from the quadrature point that is symmetric with respect to the
! plane half way between v1 and v2.  Second, the transformation from a given
! tetrahedron to the canonical tetrahedron on which the basis functions are
! tabled must also include the reordering of the vertices.  This cannot be done
! through a composition of the operators via operations on the transformation
! matrix and permutation matrix, because the permutation matrix does not apply
! uniformly over the tetrahedron.  The transformation matrix and its inverse
! must be recomputed for each edge and face as the basis functions for them are
! computed.
! The permutation of the quadrature points for each edge and face was
! determined above by calling set_quadrature_point_permutation.

! We begin by computing the usual transformation matrix, because the vertex
! basis functions are not permuted.

if (needA) then

   Ainv(1,1) = xvert(4)-xvert(1)
   Ainv(1,2) = xvert(3)-xvert(1)
   Ainv(1,3) = xvert(2)-xvert(1)
   Ainv(2,1) = yvert(4)-yvert(1)
   Ainv(2,2) = yvert(3)-yvert(1)
   Ainv(2,3) = yvert(2)-yvert(1)
   Ainv(3,1) = zvert(4)-zvert(1)
   Ainv(3,2) = zvert(3)-zvert(1)
   Ainv(3,3) = zvert(2)-zvert(1)

   denom = Ainv(1,1)*(Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)) - &
           Ainv(2,1)*(Ainv(1,2)*Ainv(3,3)-Ainv(3,2)*Ainv(1,3)) + &
           Ainv(3,1)*(Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3))

   A(1,1) = Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)
   A(1,2) = Ainv(3,2)*Ainv(1,3)-Ainv(1,2)*Ainv(3,3)
   A(1,3) = Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3)
   A(2,1) = Ainv(3,1)*Ainv(2,3)-Ainv(2,1)*Ainv(3,3)
   A(2,2) = Ainv(1,1)*Ainv(3,3)-Ainv(3,1)*Ainv(1,3)
   A(2,3) = Ainv(2,1)*Ainv(1,3)-Ainv(1,1)*Ainv(2,3)
   A(3,1) = Ainv(2,1)*Ainv(3,2)-Ainv(3,1)*Ainv(2,2)
   A(3,2) = Ainv(3,1)*Ainv(1,2)-Ainv(1,1)*Ainv(3,2)
   A(3,3) = Ainv(1,1)*Ainv(2,2)-Ainv(2,1)*Ainv(1,2)

   A = A/denom

endif

! vertex basis functions

if (present(basis)) then
   basis (1:4,:) = basis_table_tet(order)%basis(1:4,:)
endif
if (present(basisx)) then
   basisx(1:4,:) = A(1,1)*basis_table_tet(order)%basisx(1:4,:) + &
                   A(2,1)*basis_table_tet(order)%basisy(1:4,:) + &
                   A(3,1)*basis_table_tet(order)%basisz(1:4,:)
endif
if (present(basisy)) then
   basisy(1:4,:) = A(1,2)*basis_table_tet(order)%basisx(1:4,:) + &
                   A(2,2)*basis_table_tet(order)%basisy(1:4,:) + &
                   A(3,2)*basis_table_tet(order)%basisz(1:4,:)
endif
if (present(basisz)) then
   basisz(1:4,:) = A(1,3)*basis_table_tet(order)%basisx(1:4,:) + &
                   A(2,3)*basis_table_tet(order)%basisy(1:4,:) + &
                   A(3,3)*basis_table_tet(order)%basisz(1:4,:)
endif

! edge basis functions

isub = 5 ! next position in the returned basis
jnum = 0
do edge=1,6
   if (degree(edge) == 1) cycle
   jstart = 5 + (edge-1)*(deg-1) ! beginning of this edge in the basis table

! check_edge_vertex_swap returns true if the vertices of this edge are swapped
! when computing the basis function, and it sets v1 through v4 to the
! reordering of the vertices

   swapped_edge_vert = check_edge_vertex_swap(edge)

! recompute the transformation matrix with the permutation operator

   if (needA) then
      Ainv(1,1) = xvert(v4)-xvert(v1)
      Ainv(1,2) = xvert(v3)-xvert(v1)
      Ainv(1,3) = xvert(v2)-xvert(v1)
      Ainv(2,1) = yvert(v4)-yvert(v1)
      Ainv(2,2) = yvert(v3)-yvert(v1)
      Ainv(2,3) = yvert(v2)-yvert(v1)
      Ainv(3,1) = zvert(v4)-zvert(v1)
      Ainv(3,2) = zvert(v3)-zvert(v1)
      Ainv(3,3) = zvert(v2)-zvert(v1)
      denom = Ainv(1,1)*(Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)) - &
              Ainv(2,1)*(Ainv(1,2)*Ainv(3,3)-Ainv(3,2)*Ainv(1,3)) + &
              Ainv(3,1)*(Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3))
      A(1,1) = Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)
      A(1,2) = Ainv(3,2)*Ainv(1,3)-Ainv(1,2)*Ainv(3,3)
      A(1,3) = Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3)
      A(2,1) = Ainv(3,1)*Ainv(2,3)-Ainv(2,1)*Ainv(3,3)
      A(2,2) = Ainv(1,1)*Ainv(3,3)-Ainv(3,1)*Ainv(1,3)
      A(2,3) = Ainv(2,1)*Ainv(1,3)-Ainv(1,1)*Ainv(2,3)
      A(3,1) = Ainv(2,1)*Ainv(3,2)-Ainv(3,1)*Ainv(2,2)
      A(3,2) = Ainv(3,1)*Ainv(1,2)-Ainv(1,1)*Ainv(3,2)
      A(3,3) = Ainv(1,1)*Ainv(2,2)-Ainv(2,1)*Ainv(1,2)
      A = A/denom
   endif

   jnum = degree(edge)-1

   if (present(basis)) then
      if (swapped_edge_vert) then
         basis (isub:isub+jnum-1,:) = &
            basis_table_tet(order)%basis(jstart:jstart+jnum-1, &
                                         qp_perm_edge(order)%perm(:,edge))
      else
         basis (isub:isub+jnum-1,:) = &
            basis_table_tet(order)%basis(jstart:jstart+jnum-1,:)
      endif
   endif

   if (present(basisx)) then
      if (swapped_edge_vert) then
         basisx(isub:isub+jnum-1,:) = &
            A(1,1)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge)) + &
            A(2,1)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge)) + &
            A(3,1)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge))
      else
         basisx(isub:isub+jnum-1,:) = &
            A(1,1)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
            A(2,1)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
            A(3,1)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
      endif
   endif

   if (present(basisy)) then
      if (swapped_edge_vert) then
         basisy(isub:isub+jnum-1,:) = &
            A(1,2)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge)) + &
            A(2,2)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge)) + &
            A(3,2)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge))
      else
         basisy(isub:isub+jnum-1,:) = &
            A(1,2)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
            A(2,2)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
            A(3,2)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
      endif
   endif

   if (present(basisz)) then
      if (swapped_edge_vert) then
         basisz(isub:isub+jnum-1,:) = &
            A(1,3)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge)) + &
            A(2,3)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge)) + &
            A(3,3)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1, &
                                          qp_perm_edge(order)%perm(:,edge))
      else
         basisz(isub:isub+jnum-1,:) = &
            A(1,3)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
            A(2,3)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
            A(3,3)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
      endif
   endif

   isub = isub+jnum

end do

! face basis functions

do face=1,4
   if (degree(6+face) < 3) cycle
   jstart = 5 + 6*(deg-1) + ((face-1)*(deg-1)*(deg-2))/2

! check_face_vertex_swap returns an integer 1-6 indicating what permutation of
! the vertices of this face was used when computing the basis function, and it
! sets v1 through v4 to the reordering of the vertices.  A value of 6 indicates
! the vertices were not permuted.

   swapped_face_vert = check_face_vertex_swap(face)

   if (needA) then
      Ainv(1,1) = xvert(v4)-xvert(v1)
      Ainv(1,2) = xvert(v3)-xvert(v1)
      Ainv(1,3) = xvert(v2)-xvert(v1)
      Ainv(2,1) = yvert(v4)-yvert(v1)
      Ainv(2,2) = yvert(v3)-yvert(v1)
      Ainv(2,3) = yvert(v2)-yvert(v1)
      Ainv(3,1) = zvert(v4)-zvert(v1)
      Ainv(3,2) = zvert(v3)-zvert(v1)
      Ainv(3,3) = zvert(v2)-zvert(v1)
      denom = Ainv(1,1)*(Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)) - &
              Ainv(2,1)*(Ainv(1,2)*Ainv(3,3)-Ainv(3,2)*Ainv(1,3)) + &
              Ainv(3,1)*(Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3))
      A(1,1) = Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)
      A(1,2) = Ainv(3,2)*Ainv(1,3)-Ainv(1,2)*Ainv(3,3)
      A(1,3) = Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3)
      A(2,1) = Ainv(3,1)*Ainv(2,3)-Ainv(2,1)*Ainv(3,3)
      A(2,2) = Ainv(1,1)*Ainv(3,3)-Ainv(3,1)*Ainv(1,3)
      A(2,3) = Ainv(2,1)*Ainv(1,3)-Ainv(1,1)*Ainv(2,3)
      A(3,1) = Ainv(2,1)*Ainv(3,2)-Ainv(3,1)*Ainv(2,2)
      A(3,2) = Ainv(3,1)*Ainv(1,2)-Ainv(1,1)*Ainv(3,2)
      A(3,3) = Ainv(1,1)*Ainv(2,2)-Ainv(2,1)*Ainv(1,2)
      A = A/denom
   endif

   jnum = ((degree(6+face)-1)*(degree(6+face)-2))/2

   if (present(basis)) then
      if (swapped_face_vert == 6) then
         basis (isub:isub+jnum-1,:) = &
            basis_table_tet(order)%basis(jstart:jstart+jnum-1,:)
      else
         basis (isub:isub+jnum-1,:) = &
            basis_table_tet(order)%basis(jstart:jstart+jnum-1, &
                             qp_perm_face(order)%perm(:,swapped_face_vert,face))
      endif
   endif

   if (present(basisx)) then
      if (swapped_face_vert == 6) then
         basisx(isub:isub+jnum-1,:) = &
            A(1,1)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
            A(2,1)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
            A(3,1)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
      else
         basisx(isub:isub+jnum-1,:) = &
            A(1,1)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face)) + &
            A(2,1)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face)) + &
            A(3,1)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face))
      endif
   endif

   if (present(basisy)) then
      if (swapped_face_vert == 6) then
         basisy(isub:isub+jnum-1,:) = &
            A(1,2)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
            A(2,2)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
            A(3,2)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
      else
         basisy(isub:isub+jnum-1,:) = &
            A(1,2)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face)) + &
            A(2,2)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face)) + &
            A(3,2)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face))
      endif
   endif

   if (present(basisz)) then
      if (swapped_face_vert == 6) then
         basisz(isub:isub+jnum-1,:) = &
            A(1,3)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
            A(2,3)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
            A(3,3)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
      else
         basisz(isub:isub+jnum-1,:) = &
            A(1,3)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face)) + &
            A(2,3)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face)) + &
            A(3,3)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1, &
                         qp_perm_face(order)%perm(:,swapped_face_vert,face))
      endif
   endif

   isub = isub+jnum

end do

! interior basis functions

if (degree(11) >= 4) then

   jstart = 5 + 6*(deg-1) + (4*(deg-1)*(deg-2))/2
   jnum = ((degree(11)-1)*(degree(11)-2)*(degree(11)-3))/6

! the vertices are not permuted for the interior basis functions, so set the
! transformation without permutation

   if (needA) then
      Ainv(1,1) = xvert(4)-xvert(1)
      Ainv(1,2) = xvert(3)-xvert(1)
      Ainv(1,3) = xvert(2)-xvert(1)
      Ainv(2,1) = yvert(4)-yvert(1)
      Ainv(2,2) = yvert(3)-yvert(1)
      Ainv(2,3) = yvert(2)-yvert(1)
      Ainv(3,1) = zvert(4)-zvert(1)
      Ainv(3,2) = zvert(3)-zvert(1)
      Ainv(3,3) = zvert(2)-zvert(1)
      denom = Ainv(1,1)*(Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)) - &
              Ainv(2,1)*(Ainv(1,2)*Ainv(3,3)-Ainv(3,2)*Ainv(1,3)) + &
              Ainv(3,1)*(Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3))
      A(1,1) = Ainv(2,2)*Ainv(3,3)-Ainv(3,2)*Ainv(2,3)
      A(1,2) = Ainv(3,2)*Ainv(1,3)-Ainv(1,2)*Ainv(3,3)
      A(1,3) = Ainv(1,2)*Ainv(2,3)-Ainv(2,2)*Ainv(1,3)
      A(2,1) = Ainv(3,1)*Ainv(2,3)-Ainv(2,1)*Ainv(3,3)
      A(2,2) = Ainv(1,1)*Ainv(3,3)-Ainv(3,1)*Ainv(1,3)
      A(2,3) = Ainv(2,1)*Ainv(1,3)-Ainv(1,1)*Ainv(2,3)
      A(3,1) = Ainv(2,1)*Ainv(3,2)-Ainv(3,1)*Ainv(2,2)
      A(3,2) = Ainv(3,1)*Ainv(1,2)-Ainv(1,1)*Ainv(3,2)
      A(3,3) = Ainv(1,1)*Ainv(2,2)-Ainv(2,1)*Ainv(1,2)
      A = A/denom
   endif

   if (present(basis)) then
   basis (isub:isub+jnum-1,:) = &
      basis_table_tet(order)%basis(jstart:jstart+jnum-1,:)
   endif

   if (present(basisx)) then
      basisx(isub:isub+jnum-1,:) = &
         A(1,1)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
         A(2,1)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
         A(3,1)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
   endif

   if (present(basisy)) then
      basisy(isub:isub+jnum-1,:) = &
         A(1,2)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
         A(2,2)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
         A(3,2)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
   endif

   if (present(basisz)) then
      basisz(isub:isub+jnum-1,:) = &
         A(1,3)*basis_table_tet(order)%basisx(jstart:jstart+jnum-1,:) + &
         A(2,3)*basis_table_tet(order)%basisy(jstart:jstart+jnum-1,:) + &
         A(3,3)*basis_table_tet(order)%basisz(jstart:jstart+jnum-1,:)
   endif

endif

nbasis = isub+jnum-1

! Return the quadrature rule if requested

if (present(nqpoints)) then
   call quadrature_rule_tet(order,xvert,yvert,zvert,nq,qw,xq,yq,zq,jerr)
   if (jerr /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("p_hier_basis_at_quad_pts_tet: quadrature_rule_tet returned error code",intlist=(/jerr/))
      stop
   endif
   allocate(qweights(nq),xquad(nq),yquad(nq),zquad(nq),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("p_hier_basis_at_quad_pts_tet: allocation failed")
      stop
   endif
   nqpoints = nq
   qweights = qw
   xquad = xq
   yquad = yq
   zquad = zq
   deallocate(qw,xq,yq,zq)
endif


contains

!        ----------------------
function check_edge_vertex_swap(edge)
!        ----------------------

!----------------------------------------------------
! This routine determines if the vertices of edge edge get swapped in
! p_hier_basis_func_tet.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: edge
logical :: check_edge_vertex_swap
!----------------------------------------------------
! Local variables:

integer :: vertex1, vertex2, holdvertex1
!----------------------------------------------------
! Begin executable code

! identify the endpoints of the edge in the canonial edge ordering

if (edge < 4) then
   vertex1 = 1 + mod(edge,3)
   vertex2 = 1 + mod(edge+1,3)
else
   vertex1 = edge-3
   vertex2 = 4
endif
if (edge == 2) then
   vertex1 = 1
   vertex2 = 3
endif
holdvertex1 = vertex1

! sort them, just like in module basis_functions

call sort_verts(vertex1,vertex2,xvert,yvert,zvert)

! they are swapped only if vertex1 changed

check_edge_vertex_swap = vertex1/=holdvertex1

select case(edge)
case(1)
   v1 = 1; v2 = vertex1; v3 = vertex2; v4 = 4
case(2)
   v1 = vertex1; v2 = 2; v3 = vertex2; v4 = 4
case(3)
   v1 = vertex1; v2 = vertex2; v3 = 3; v4 = 4
case(4)
   v1 = vertex1; v2 = 2; v3 = 3; v4 = vertex2
case(5)
   v1 = 1; v2 = vertex1; v3 = 3; v4 = vertex2
case(6)
   v1 = 1; v2 = 2; v3 = vertex1; v4 = vertex2
end select

end function check_edge_vertex_swap

!        ----------------------
function check_face_vertex_swap(face)
!        ----------------------

!----------------------------------------------------
! This routine determines how the vertices of face face get swapped in
! p_hier_basis_func_tet.
! 1 - vertices 1 and 2 are swapped
! 2 - vertices 2 and 3 are swapped
! 3 - vertices 3 and 1 are swapped
! 4 - rotated 1->2, 2->3 and 3->1
! 5 - rotated 1->3, 2->1 and 3->2
! 6 - not swapped
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: face
integer :: check_face_vertex_swap
!----------------------------------------------------
! Local variables:

integer :: vertex1, vertex2, vertex3, holdvertex1, holdvertex2
!----------------------------------------------------
! Begin executable code

! identify the vertices of the face

vertex1 = 1 + mod(face,4)
vertex2 = 1 + mod(vertex1,4)
vertex3 = 1 + mod(vertex2,4)
if (face == 2) then
   vertex1 = 1
   vertex2 = 3
   vertex3 = 4
endif
if (face == 3) then
   vertex1 = 1
   vertex2 = 2
   vertex3 = 4
endif

holdvertex1 = vertex1
holdvertex2 = vertex2

! sort them, just like in module basis_functions

call sort_verts(vertex1,vertex2,xvert,yvert,zvert)
call sort_verts(vertex2,vertex3,xvert,yvert,zvert)
call sort_verts(vertex1,vertex2,xvert,yvert,zvert)

! determine how they were swapped

if (vertex1 == holdvertex1) then
   if (vertex2 == holdvertex2) then
      check_face_vertex_swap = 6
   else
      check_face_vertex_swap = 2
   endif
elseif (vertex1 == holdvertex2) then
   if (vertex2 == holdvertex1) then
      check_face_vertex_swap = 1
   else
      check_face_vertex_swap = 4
   endif
else ! vertex1 == holdvertex3
   if (vertex3 == holdvertex1) then
      check_face_vertex_swap = 3
   else
      check_face_vertex_swap = 5
   endif
endif

select case(face)
case(1)
   v1 = 1; v2 = vertex1; v3 = vertex2; v4 = vertex3
case(2)
   v1 = vertex1; v2 = 2; v3 = vertex2; v4 = vertex3
case(3)
   v1 = vertex1; v2 = vertex2; v3 = 3; v4 = vertex3
case(4)
   v1 = vertex1; v2 = vertex2; v3 = vertex3; v4 = 4
end select

end function check_face_vertex_swap

end subroutine p_hier_basis_at_quad_pts_tet

!          --------------------------------
subroutine set_quadrature_point_permutation(order)
!          --------------------------------

!----------------------------------------------------
! This routine sets the permutation of quadrature points of the quadrature rule
! of order order to be used when the order of the vertices is changed by
! sorting.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: order

!----------------------------------------------------
! Parameters

integer, parameter :: canonical1(6) = (/2,1,1,1,2,3/), &
                      canonical2(6) = (/3,3,2,4,4,4/), &
                      NotOnLine1(6) = (/1,2,3,2,1,1/), &
                      NotOnLine2(6) = (/4,4,4,3,3,2/)

!----------------------------------------------------
! Local variables:

integer :: edge, face, v1, v2, v3, v4, w1, w2, w3, w4, loc_order, nqpoints, &
           astat, i, j, qp, swapqp, swapqp2, swaptype
real(my_real), pointer :: bary(:,:)

!----------------------------------------------------
! Begin executable code

! The basic premise:  For the edge (read also face) basis functions for a
! tetrahedron, the vertices are sorted by size of the coordinates so that they
! are in the same order for every element that contains that edge, required for
! continuity of the basis function.

! The tetrahedron quadrature rules are only defined for odd orders.  Even orders
! use the rule for order+1.

if (2*(order/2) == order) then
   loc_order = order+1
else
   loc_order = order
endif

! Get the barycentric coordinates of the quadrature points.
! bary(:,1) contains the alpha used in defining the quadrature points,
! bary(:,2) contains beta, etc. alpha, beta, gamma and delta contain the
! weight of vertex 1, 2, 3 and 4 (in local numbering) for each quadrature point.

call get_quadrature_point_bary(order,bary)
nqpoints = size(bary,dim=1)

! Allocate the permutation vectors

allocate(qp_perm_edge(loc_order)%perm(nqpoints,6), &
         qp_perm_face(loc_order)%perm(nqpoints,5,4),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("set_quadrature_point_permutation: allocation failed")
   stop
endif

! Define permutation vectors for the edges.

do edge=1,6

! Get the local index of the vertices of this edge, in the canonical order used
! in module basis_functions.

   v1 = canonical1(edge)
   v2 = canonical2(edge)
   v3 = NotOnLine1(edge)
   v4 = NotOnLine2(edge)

! Begin with the identity permutation.

   qp_perm_edge(loc_order)%perm(:,edge) = (/(i,i=1,nqpoints)/)

! For each quadrature point ...

   do qp=1,nqpoints-1

! If this entry of perm is not the identity, this point has already been swapped

      if (qp_perm_edge(loc_order)%perm(qp,edge) /= qp) cycle

! If the weights of the two vertices are the same, the point lies on the plane
! half way between them, so there is nothing to swap.

      if (fuzzy_equal(bary(qp,v1),bary(qp,v2))) cycle

! Otherwise, look for the point later in the list that has the weights of v1
! and v2 swapped and the weights of each of the other two vertices the same.
! This is the point that is symmetric to the current quadrature point with
! respect to these two vertices.

      swapqp = -1
      do j=qp+1,nqpoints
         if (fuzzy_equal(bary(j,v1),bary(qp,v2)) .and. &
             fuzzy_equal(bary(j,v2),bary(qp,v1)) .and. &
             fuzzy_equal(bary(j,v3),bary(qp,v3)) .and. &
             fuzzy_equal(bary(j,v4),bary(qp,v4))) then
            swapqp = j
            exit
         endif
      end do

      if (swapqp == -1) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("set_quadrature_point_permutation: could not find symmetric quadrature point")
         stop
      endif

! Swap these two points.

      qp_perm_edge(loc_order)%perm(qp,edge) = swapqp
      qp_perm_edge(loc_order)%perm(swapqp,edge) = qp

   end do ! qp
end do ! edge

! Define permutation vectors for the faces.

do face=1,4

! Get the local index of the vertices of this face, in the same order used
! in module basis_functions.

   v1 = 1 + mod(face,4)
   v2 = 1 + mod(v1,4)
   v3 = 1 + mod(v2,4)
   v4 = 1 + mod(v3,4)
   if (face == 2) then
      v1 = 1
      v2 = 3
      v3 = 4
   endif
   if (face == 3) then
      v1 = 1
      v2 = 2
      v3 = 4
   endif

! There are 5 types of vertex swaps for a face, as described in
! check_face_vertex_swap.  The first 3 swap two vertices, so their permutations
! are defined just like those for the edges, with the appropriate selection
! of vertices.

   do swaptype=1,3

      select case(swaptype)
      case(1) ! vertices 1 and 2 are swapped
         w1=v1; w2=v2; w3=v3; w4=v4
      case(2) ! vertices 2 and 3 are swapped
         w1=v2; w2=v3; w3=v1; w4=v4
      case(3) ! vertices 3 and 1 are swapped
         w1=v3; w2=v1; w3=v2; w4=v4
      end select

      qp_perm_face(loc_order)%perm(:,swaptype,face) = (/(i,i=1,nqpoints)/)

      do qp=1,nqpoints-1

         if (qp_perm_face(loc_order)%perm(qp,swaptype,face) /= qp) cycle

         if (fuzzy_equal(bary(qp,w1),bary(qp,w2))) cycle

         swapqp = -1
         do j=qp+1,nqpoints
            if (fuzzy_equal(bary(j,w1),bary(qp,w2)) .and. &
                fuzzy_equal(bary(j,w2),bary(qp,w1)) .and. &
                fuzzy_equal(bary(j,w3),bary(qp,w3)) .and. &
                fuzzy_equal(bary(j,w4),bary(qp,w4))) then
               swapqp = j
               exit
            endif
         end do

         if (swapqp == -1) then
            ierr = PHAML_INTERNAL_ERROR
            call fatal("set_quadrature_point_permutation: could not find symmetric quadrature point")
            stop
         endif

         qp_perm_face(loc_order)%perm(qp,swaptype,face) = swapqp
         qp_perm_face(loc_order)%perm(swapqp,swaptype,face) = qp

      end do ! qp
   end do ! swaptype

! The other 2 rotate three vertices and can be handled similarly, except there
! are three required movements of the barycentric coordinates.  The same points
! are used for both type 4 and type 5, but they are rotated in opposite
! directions.

   qp_perm_face(loc_order)%perm(:,4,face) = (/(i,i=1,nqpoints)/)
   qp_perm_face(loc_order)%perm(:,5,face) = (/(i,i=1,nqpoints)/)

   do qp=1,nqpoints-1

      if (qp_perm_face(loc_order)%perm(qp,4,face) /= qp) cycle

! If the weights of all three vertices are the same, the point lies on the line
! that is the center of rotation, so there is nothing to swap.

      if (fuzzy_equal(bary(qp,v1),bary(qp,v2)) .and. &
          fuzzy_equal(bary(qp,v1),bary(qp,v3))) cycle

! Find two swap points, one that is shifted once and one that is shifted twice.

      swapqp = -1
      swapqp2 = -1
      do j=qp+1,nqpoints
         if (fuzzy_equal(bary(j,v2),bary(qp,v1)) .and. &
             fuzzy_equal(bary(j,v3),bary(qp,v2)) .and. &
             fuzzy_equal(bary(j,v1),bary(qp,v3))) then
            swapqp = j
            if (swapqp2 /= -1) exit
         endif
         if (fuzzy_equal(bary(j,v3),bary(qp,v1)) .and. &
             fuzzy_equal(bary(j,v1),bary(qp,v2)) .and. &
             fuzzy_equal(bary(j,v2),bary(qp,v3))) then
            swapqp2 = j
            if (swapqp /= -1) exit
         endif
      end do

      if (swapqp == -1 .or. swapqp2 == -1) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("set_quadrature_point_permutation: could not find symmetric rotation point")
         stop
      endif

! Perform the rotations.

! Rotate 1->2, 2->3 and 3->1

      qp_perm_face(loc_order)%perm(swapqp,4,face) = qp
      qp_perm_face(loc_order)%perm(swapqp2,4,face) = swapqp
      qp_perm_face(loc_order)%perm(qp,4,face) = swapqp2

! Rotate 1->3, 2->1 and 3->2

      qp_perm_face(loc_order)%perm(swapqp2,5,face) = qp
      qp_perm_face(loc_order)%perm(qp,5,face) = swapqp
      qp_perm_face(loc_order)%perm(swapqp,5,face) = swapqp2

   end do ! qp

end do ! face

! Copy the permutations to loc_order-1, which is the even order that uses the
! same quadrature rule.

if (loc_order /= 1) then
   allocate(qp_perm_edge(loc_order-1)%perm(nqpoints,6), &
            qp_perm_face(loc_order-1)%perm(nqpoints,5,4),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("set_quadrature_point_permutation: allocation failed")
      stop
   endif
   qp_perm_edge(loc_order-1)%perm = qp_perm_edge(loc_order)%perm
   qp_perm_face(loc_order-1)%perm = qp_perm_face(loc_order)%perm
endif

! Free memory.

if (associated(bary)) deallocate(bary)

contains

function fuzzy_equal(real1,real2)
real(my_real), intent(in) :: real1,real2
logical :: fuzzy_equal
real(my_real), parameter :: eps = 100*epsilon(0.0_my_real)
fuzzy_equal = abs(real1-real2)/max(abs(real1),abs(real2)) < eps
end function fuzzy_equal

end subroutine set_quadrature_point_permutation

end module basis_at_quad_pts
