!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

!----------------------------------------------------
! This file contains the user supplied external subroutines that define
! the PDE(s) to be solved, and other required external subroutines.
!   pdecoefs bconds boundary_point boundary_npiece boundary_param iconds trues
!   truexs trueys update_usermod phaml_integral_kernel
!
! This version illustrates a system of elliptic PDEs solved simultaneously.
!----------------------------------------------------

!          --------
subroutine pdecoefs(x,y,z,cxx,cyy,czz,cxy,cxz,cyz,cx,cy,cz,c,rs)
!          --------

!----------------------------------------------------
! This subroutine returns the coefficient and right hand side of the PDEs
! at the points (x,y)
!
! PDEs are
!
!    -( cxx(x,y) * u  )  -( cyy(x,y) * u  ) + c(x,y) * u = rs(x,y)
!                   x  x                y  y
!
! where cxx, cyy and c are matrices in which the (i,j)th entry is the
! coefficient of the jth solution in the ith equation, u is a vector of
! the solutions, and rs is a vector of the right hand sides.
!
! For eigenvalue problems, the right hand side is lambda * u * rs(x,y)
!
! cxy, cx and cy are not currently used and should not be set.  They are
! for future expansion.
!
! NOTE: BE CAREFUL TO GET THE SIGNS RIGHT
! e.g. cxx=cyy=1 means rs=-(uxx+uyy)
!
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
real(my_real), intent(out) :: cxx(:,:),cyy(:,:),czz(:,:),cxy(:,:),cxz(:,:), &
                              cyz(:,:),cx(:,:),cy(:,:),cz(:,:),c(:,:),rs(:)
!----------------------------------------------------

!----------------------------------------------------
! Local variables

real(my_real), parameter :: pi = 3.14159265358979323844_my_real
!----------------------------------------------------

!----------------------------------------------------
! Begin executable code

if (size(cxx,dim=1) == 2) then

! System of 2 equations

! -uxx -uyy -uzz + v = f1
! -vxx -vyy -vzz + u = f2

   cxy=0; cxz=0; cyz=0; cx=0; cy=0; cz=0

   cxx(1,1) = 1.0_my_real;  cxx(1,2) = 0.0_my_real
   cxx(2,1) = 0.0_my_real;  cxx(2,2) = 1.0_my_real

   cyy(1,1) = 1.0_my_real;  cyy(1,2) = 0.0_my_real
   cyy(2,1) = 0.0_my_real;  cyy(2,2) = 1.0_my_real

   czz(1,1) = 1.0_my_real;  czz(1,2) = 0.0_my_real
   czz(2,1) = 0.0_my_real;  czz(2,2) = 1.0_my_real

     c(1,1) = 0.0_my_real;    c(1,2) = 1.0_my_real
     c(2,1) = 1.0_my_real;    c(2,2) = 0.0_my_real

   rs(1) = -(2.0_my_real*exp(x-y) - (x+y)**4/8.0_my_real)
   rs(2) = -(3.0_my_real*(x+y)**2 - exp(x-y))

else

! System of 3 equations

! -uxx -uyy -uzz + v -w = f1
! -vxx -vyy -vzz + u    = f2
! -wxx -wyy -wzz + 8*v  = f3

   cxy=0; cxz=0; cyz=0; cx=0; cy=0; cz=0

   cxx(1,1) = 1.0_my_real;  cxx(1,2) = 0.0_my_real;  cxx(1,3) = 0.0_my_real
   cxx(2,1) = 0.0_my_real;  cxx(2,2) = 1.0_my_real;  cxx(2,3) = 0.0_my_real
   cxx(3,1) = 0.0_my_real;  cxx(3,2) = 0.0_my_real;  cxx(3,3) = 1.0_my_real

   cyy(1,1) = 1.0_my_real;  cyy(1,2) = 0.0_my_real;  cyy(1,3) = 0.0_my_real
   cyy(2,1) = 0.0_my_real;  cyy(2,2) = 1.0_my_real;  cyy(2,3) = 0.0_my_real
   cyy(3,1) = 0.0_my_real;  cyy(3,2) = 0.0_my_real;  cyy(3,3) = 1.0_my_real

   czz(1,1) = 1.0_my_real;  czz(1,2) = 0.0_my_real;  czz(1,3) = 0.0_my_real
   czz(2,1) = 0.0_my_real;  czz(2,2) = 1.0_my_real;  czz(2,3) = 0.0_my_real
   czz(3,1) = 0.0_my_real;  czz(3,2) = 0.0_my_real;  czz(3,3) = 1.0_my_real

     c(1,1) = 0.0_my_real;    c(1,2) = 1.0_my_real;    c(1,3) =-1.0_my_real
     c(2,1) = 1.0_my_real;    c(2,2) = 0.0_my_real;    c(2,3) = 0.0_my_real
     c(3,1) = 0.0_my_real;    c(3,2) = 8.0_my_real;    c(3,3) = 0.0_my_real

   rs(1) = -(2.0_my_real*exp(x-y) - (x+y)**4/8.0_my_real + sin(pi*x)*sin(pi*y)*sin(pi*z))
   rs(2) = -(3.0_my_real*(x+y)**2 - exp(x-y))
   rs(3) = -(-3*pi**2*sin(pi*x)*sin(pi*y)*sin(pi*z) - (x+y)**4)

endif

end subroutine pdecoefs

!          -----
subroutine bconds(x,y,z,bmark,itype,c,rs)
!          -----

!----------------------------------------------------
! This subroutine returns the boundary conditions at the points (x,y).
!
! Each boundary condition is either
!
!    u = rs(x,y) or  u  + c(x,y)*u = rs(x,y)
!                     n
!
! c is an array in which the (i,j)th component is the coefficient of the
! jth solution in the ith boundary condition.  rs is a vector in which the
! ith component is the right side of the ith boundary condition.
!
! itype indicates which type of condition applies to each point, using
! symbolic constants from module phaml.  It is DIRICHLET for the first
! condition, NATURAL for the second condition with c==0, and MIXED for
! the second condition with c/=0.  The ith component is the type of the
! ith boundary condition.
!
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
integer, intent(in) :: bmark
integer, intent(out) :: itype(:)
real(my_real), intent(out) :: c(:,:),rs(:)
!----------------------------------------------------

!----------------------------------------------------
! Non-module procedures used are:

interface

   function trues(x,y,z,comp,eigen) ! real (my_real)
   use phaml
   real (my_real), intent(in) :: x,y,z
   integer, intent(in) :: comp,eigen
   real (my_real) :: trues
   end function trues

end interface

!----------------------------------------------------
! Local variables

!----------------------------------------------------

!----------------------------------------------------
! Begin executable code

! Dirichlet boundary conditions

itype = DIRICHLET
c = 0.0_my_real
rs(1) = trues(x,y,z,1,1)
rs(2) = trues(x,y,z,2,1)
if (size(rs) == 3) then
   rs(3) = trues(x,y,z,3,1)
endif

end subroutine bconds

!        ------
function iconds(x,y,z,comp,eigen)
!        ------

!----------------------------------------------------
! This routine returns the initial condition of the comp'th solution
! for a time dependent problem.
! It can also be used for the initial guess for a nonlinear problem, or
! systems of equations solved by sucessive substitution.
! For problems where there are no initial conditions, it is a dummy.
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
integer, intent(in) :: comp,eigen
real(my_real) :: iconds

!----------------------------------------------------
! Begin executable code

iconds = 0.0_my_real

end function iconds

!        -----
function trues(x,y,z,comp,eigen) ! real (my_real)
!        -----

!----------------------------------------------------
! This is the comp'th true solution of the differential equation, if known.
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
integer, intent(in) :: comp,eigen
real (my_real) :: trues
!----------------------------------------------------

real(my_real), parameter :: pi = 3.14159265358979323844_my_real
!----------------------------------------------------
! Begin executable code

select case (comp)

   case (1)
      trues = exp(x-y)

   case (2)
      trues = (x+y)**4/8.0_my_real

   case (3)
      trues = sin(pi*x)*sin(pi*y)*sin(pi*z)

   case default
      print *,"ERROR -- bad value ",comp," for comp in trues"
      stop

end select

end function trues

!        ------
function truexs(x,y,z,comp,eigen) ! real (my_real)
!        ------

!----------------------------------------------------
! This is the x derivative of the comp'th true solution of the differential
! equation, if known.
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
integer, intent(in) :: comp,eigen
real (my_real) :: truexs
!----------------------------------------------------

real(my_real), parameter :: pi = 3.14159265358979323844_my_real
!----------------------------------------------------
! Begin executable code

select case (comp)

   case (1)
      truexs = exp(x-y)

   case (2)
      truexs = (x+y)**3/2.0_my_real

   case (3)
      truexs = pi*cos(pi*x)*sin(pi*y)*sin(pi*z)

   case default
      print *,"ERROR -- bad value ",comp," for comp in truexs"
      stop

end select

end function truexs

!        ------
function trueys(x,y,z,comp,eigen) ! real (my_real)
!        ------

!----------------------------------------------------
! This is the y derivative of the comp'th true solution of the differential
! equation, if known.
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
integer, intent(in) :: comp,eigen
real (my_real) :: trueys
!----------------------------------------------------

real(my_real), parameter :: pi = 3.14159265358979323844_my_real
!----------------------------------------------------
! Begin executable code

select case (comp)

   case (1)
      trueys = -exp(x-y)

   case (2)
      trueys = (x+y)**3/2.0_my_real

   case (3)
      trueys = pi*sin(pi*x)*cos(pi*y)*sin(pi*z)

   case default
      print *,"ERROR -- bad value ",comp," for comp in trueys"
      stop

end select

end function trueys

!        ------
function truezs(x,y,z,comp,eigen) ! real (my_real)
!        ------

!----------------------------------------------------
! This is the z derivative of the comp'th true solution of the differential
! equation, if known.
!----------------------------------------------------

!----------------------------------------------------
! Modules used are:

use phaml
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

real(my_real), intent(in) :: x,y,z
integer, intent(in) :: comp,eigen
real (my_real) :: truezs
!----------------------------------------------------

real(my_real), parameter :: pi = 3.14159265358979323844_my_real
!----------------------------------------------------
! Begin executable code

select case (comp)

   case (1)
      truezs = 0

   case (2)
      truezs = 0

   case (3)
      truezs = pi*sin(pi*x)*sin(pi*y)*cos(pi*z)

   case default
      print *,"ERROR -- bad value ",comp," for comp in truezs"
      stop

end select

end function truezs

!          --------------
subroutine update_usermod(phaml_solution)
!          --------------

use phaml
type(phaml_solution_type), intent(in) :: phaml_solution

! Dummy version

end subroutine update_usermod

!        ---------------------
function phaml_integral_kernel(kernel,x,y,z)
!        ---------------------

use phaml
integer, intent(in) :: kernel
real(my_real), intent(in) :: x,y,z
real(my_real) :: phaml_integral_kernel

! Identity function

phaml_integral_kernel = 1.0

end function phaml_integral_kernel

!        ----------
function regularity(x,y,z)
!        ----------

use phaml
real(my_real), intent(in) :: x(*),y(*),z(*)
real(my_real) :: regularity

! Dummy version, assume infinitely differentiable everywhere.

regularity = huge(0.0_my_real)

end function regularity
